import pytest
import torch
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor import format
from dmx.compressor.numerical.observer import MinMaxObserver, HistogramObserver
from dmx.compressor.numerical import CastTo
from dmx.compressor.advanced_recipe import (
    DmxQuantizerCalibrationHyperparams,
    DmxModuleQuantizerCalibrationHyperparams,
    DmxQuantizerCalibrationRecipe,
)
from torch import nn

RANDOM_SEED = 0

torch.manual_seed(RANDOM_SEED)


BATCH_SIZE = 2
IN_DIM = 16
OUT_DIM = 8
PERTURB_CHAN = 4
KER_SIZE = 3
IMG_SIZE = 8


def _create_module(cls):
    if cls == dmxnn.Linear:
        return cls(IN_DIM, OUT_DIM)
    elif cls == dmxnn.Conv1d:
        return cls(IN_DIM, 2, KER_SIZE)
    elif cls == dmxnn.Conv2d:
        return cls(IN_DIM, OUT_DIM, KER_SIZE, KER_SIZE)
    else:
        raise ValueError("unsupported module class {cls}")


def _create_test_input(module: torch.nn.Module):
    if isinstance(module, nn.Conv1d):
        x_ref = torch.randn(BATCH_SIZE, IN_DIM, IMG_SIZE)
    elif isinstance(module, nn.Conv2d):
        x_ref = torch.randn(BATCH_SIZE, IN_DIM, IMG_SIZE, IMG_SIZE)
    else:
        x_ref = torch.randn(BATCH_SIZE, IN_DIM)

    return x_ref


def test_block_size_non_factor():
    """
    verify performance of group quant when group size is not a factor of ch_axis
    """
    cast = CastTo(
        format=format.INT4,
        observer=MinMaxObserver,
        group_size=2,
        qscheme=torch.per_tensor_symmetric,
        ch_axis=0,
    )
    cast.enable_observer()
    x = torch.Tensor([[0, 1], [3, 7], [5.1, 8], [10, 14], [0.1, 0.7]])
    y = torch.Tensor([[0, 1], [3, 7], [6, 8], [10, 14], [0.1, 0.7]])
    assert torch.allclose(cast(x), y, rtol=0.0, atol=1e-6)


def test_block_size_non_factor_linear_weight():
    """
    verify performance of group quant when group size is not a factor of ch_axis called from a linear layer
    """
    in_dim = 2
    out_dim = 5
    layer = dmxnn.Linear(in_dim, out_dim)
    layer.weight_cast.set_format(format.INT4)
    layer.weight.data = torch.Tensor([[0, 1], [3, 7], [5.1, 8], [10, 14], [0.1, 0.7]])
    hp_gen = lambda _: {
        layer: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=MinMaxObserver,
                qscheme_to_overload=torch.per_tensor_symmetric,
                group_size=2,
            ),
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(layer), torch.no_grad():
        layer._weight

    y = torch.Tensor([[0, 1], [3, 7], [6, 8], [10, 14], [0.1, 0.7]])
    assert torch.allclose(layer._weight, y, rtol=0.0, atol=1e-6)


def test_block_size_non_factor_linear_activation():
    """
    verify performance of group quant when group size is not a factor of ch_axis called from a linear layer
    """
    in_dim = 2
    out_dim = 5
    layer = dmxnn.Linear(in_dim, out_dim)
    layer.input_casts.input_cast.set_format(format.INT4)
    x = torch.Tensor([[0, 1], [3, 7], [5.1, 8], [10, 14], [0.1, 0.7]])
    layer.input_casts.input_cast.ch_axis = 0
    hp_gen = lambda _: {
        layer: DmxModuleQuantizerCalibrationHyperparams(
            inputs={
                "input_cast": DmxQuantizerCalibrationHyperparams(
                    observer_cls=MinMaxObserver,
                    qscheme_to_overload=torch.per_tensor_symmetric,
                    group_size=2,
                )
            },
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(layer), torch.no_grad():
        layer(x)

    y = torch.Tensor([[0, 1], [3, 7], [6, 8], [10, 14], [0.1, 0.7]])
    assert torch.allclose(layer.input_casts.input_cast(x), y, rtol=0.0, atol=1e-6)


def test_hypernet_linear():
    """
    Testing that applying hypernet on a column of weight matrix is equivalent to the corresponding column of the hypernet on entire weight matrix
    """
    in_dim = 2
    out_dim = 5
    layer = dmxnn.Linear(in_dim, out_dim)
    layer.weight_cast.set_format(format.INT4)
    hp_gen = lambda _: {
        layer: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=MinMaxObserver,
                qscheme_to_overload=torch.per_tensor_symmetric,
                group_size=2,
            ),
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(layer), torch.no_grad():
        layer._weight
    per_col_cast = torch.zeros((out_dim, in_dim))
    for i in range(in_dim):
        per_col_cast[:, i] = layer.weight_hypernet(layer.weight[:, i])
    assert torch.allclose(per_col_cast, layer._weight, rtol=0.0, atol=1e-6)


@pytest.mark.parametrize(
    "module_cls",
    (
        dmxnn.Linear,
        dmxnn.Conv1d,
        dmxnn.Conv2d,
    ),
)
@pytest.mark.parametrize("observer", (MinMaxObserver, HistogramObserver))
@pytest.mark.parametrize(
    "qscheme", (torch.per_tensor_affine, torch.per_tensor_symmetric)
)
@pytest.mark.parametrize("format", (format.INT4, format.INT8))
def test_per_tensor_equivalence_weight(module_cls, observer, qscheme, format):
    """
    verify that per group is equivalent to per tensor when group size is set to size of ch_axis
    """
    module = _create_module(module_cls)
    module_ref = _create_module(module_cls)
    module_ref.weight.data = module.weight.data

    module.weight_cast.set_format(format)
    module_ref.weight_cast.set_format(format)
    hp_gen = lambda _: {
        module: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=observer,
                qscheme_to_overload=qscheme,
                group_size=module.weight.shape[module.weight_cast.ch_axis],
            ),
        )
    }
    hp_gen_ref = lambda _: {
        module_ref: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=observer,
                qscheme_to_overload=qscheme,
            ),
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(module), torch.no_grad():
        module._weight
    with DmxQuantizerCalibrationRecipe(hp_gen_ref).applied_to(
        module_ref
    ), torch.no_grad():
        module_ref._weight

    assert torch.allclose(module._weight, module_ref._weight, rtol=0.0, atol=1e-8)


@pytest.mark.parametrize(
    "module_cls",
    (
        dmxnn.Linear,
        dmxnn.Conv1d,
        dmxnn.Conv2d,
    ),
)
@pytest.mark.parametrize("observer", (MinMaxObserver,))
@pytest.mark.parametrize(
    "qscheme",
    (
        [torch.per_tensor_affine, torch.per_channel_affine],
        [torch.per_tensor_symmetric, torch.per_channel_symmetric],
    ),
)
@pytest.mark.parametrize("format", (format.INT4, format.INT8))
def test_per_channel_equivalence_weight(module_cls, observer, qscheme, format):
    """
    verify that per group is equivalent to per channel when group size is set to 1
    """
    module = _create_module(module_cls)
    module_ref = _create_module(module_cls)
    module_ref.weight.data = module.weight.data

    module.weight_cast.set_format(format)
    module_ref.weight_cast.set_format(format)
    hp_gen = lambda _: {
        module: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=observer,
                qscheme_to_overload=qscheme[0],
                group_size=1,
            ),
        )
    }
    hp_gen_ref = lambda _: {
        module_ref: DmxModuleQuantizerCalibrationHyperparams(
            weight=DmxQuantizerCalibrationHyperparams(
                observer_cls=observer,
                qscheme_to_overload=qscheme[1],
            ),
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(module), torch.no_grad():
        module._weight
    with DmxQuantizerCalibrationRecipe(hp_gen_ref).applied_to(
        module_ref
    ), torch.no_grad():
        module_ref._weight

    assert torch.allclose(module._weight, module_ref._weight, rtol=0.0, atol=1e-8)


@pytest.mark.parametrize(
    "module_cls",
    (
        dmxnn.Linear,
        dmxnn.Conv1d,
        dmxnn.Conv2d,
    ),
)
@pytest.mark.parametrize("observer", (MinMaxObserver, HistogramObserver))
@pytest.mark.parametrize(
    "qscheme", (torch.per_tensor_affine, torch.per_tensor_symmetric)
)
@pytest.mark.parametrize("format", (format.INT4, format.INT8))
def test_per_tensor_equivalence_activation(module_cls, observer, qscheme, format):
    """
    verify that per group is equivalent to per tensor when group size is set to size of ch_axis
    """
    module = _create_module(module_cls)
    module_ref = _create_module(module_cls)
    x = _create_test_input(
        module,
    )
    module.input_casts.input_cast.set_format(format)
    module_ref.input_casts.input_cast.set_format(format)
    hp_gen = lambda _: {
        module: DmxModuleQuantizerCalibrationHyperparams(
            inputs={
                "input_cast": DmxQuantizerCalibrationHyperparams(
                    observer_cls=observer,
                    qscheme_to_overload=qscheme,
                    group_size=x.shape[module.input_casts.input_cast.ch_axis],
                )
            },
        )
    }
    hp_gen_ref = lambda _: {
        module_ref: DmxModuleQuantizerCalibrationHyperparams(
            inputs={
                "input_cast": DmxQuantizerCalibrationHyperparams(
                    observer_cls=observer,
                    qscheme_to_overload=qscheme,
                )
            },
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(module), torch.no_grad():
        module(x)
    with DmxQuantizerCalibrationRecipe(hp_gen_ref).applied_to(
        module_ref
    ), torch.no_grad():
        module_ref(x)
    assert torch.allclose(
        module.input_casts.input_cast(x),
        module_ref.input_casts.input_cast(x),
        rtol=0.0,
        atol=1e-8,
    )


@pytest.mark.parametrize(
    "module_cls",
    (
        dmxnn.Linear,
        dmxnn.Conv1d,
        dmxnn.Conv2d,
    ),
)
@pytest.mark.parametrize("observer", (MinMaxObserver,))
@pytest.mark.parametrize(
    "qscheme",
    (
        [torch.per_tensor_affine, torch.per_channel_affine],
        [torch.per_tensor_symmetric, torch.per_channel_symmetric],
    ),
)
@pytest.mark.parametrize("format", (format.INT4, format.INT8))
def test_per_channel_equivalence_activation(module_cls, observer, qscheme, format):
    """
    verify that per group is equivalent to per channel when group size is set to 1
    """
    module = _create_module(module_cls)
    module_ref = _create_module(module_cls)

    x = _create_test_input(
        module,
    )

    module.input_casts.input_cast.set_format(format)
    module_ref.input_casts.input_cast.set_format(format)
    hp_gen = lambda _: {
        module: DmxModuleQuantizerCalibrationHyperparams(
            inputs={
                "input_cast": DmxQuantizerCalibrationHyperparams(
                    observer_cls=observer,
                    qscheme_to_overload=qscheme[0],
                    group_size=1,
                )
            },
        )
    }
    hp_gen_ref = lambda _: {
        module_ref: DmxModuleQuantizerCalibrationHyperparams(
            inputs={
                "input_cast": DmxQuantizerCalibrationHyperparams(
                    observer_cls=observer,
                    qscheme_to_overload=qscheme[1],
                )
            },
        )
    }
    with DmxQuantizerCalibrationRecipe(hp_gen).applied_to(module), torch.no_grad():
        module(x)
    with DmxQuantizerCalibrationRecipe(hp_gen_ref).applied_to(
        module_ref
    ), torch.no_grad():
        module_ref(x)
    assert torch.allclose(
        module.input_casts.input_cast(x),
        module_ref.input_casts.input_cast(x),
        rtol=0.0,
        atol=1e-8,
    )



from typing import Any, Dict, Tuple, Union

import torch.fx as fx
from torch.fx.node import Argument, Target, map_aggregate
from torch.fx.proxy import Proxy

from .utils import *


class NodeDictTransformer(fx.Transformer):
    """
    A transformer that creates a dict contaning mapping between target of node and the node itself

    Args:
        module (fx.GraphModule): the module to create the mapping

    Attributes:
        module (fx.GraphModule): the module to create the mapping
        nodeDict (dict): dictionary to store the mapping between occurance order of the node and the node
    """

    def __init__(self, module: fx.GraphModule):
        super().__init__(module)
        self.module = module
        # A dictionary to map node.target to node instance
        self.nodeDict = dict()
        self.nodeDict["function"] = list()
        self.nodeDict["method"] = list()
        self.nodeDict["module"] = list()

    def placeholder(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``placeholder`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert isinstance(target, str)
        placeholder_node = self.new_graph.placeholder(target)
        self.nodeDict[placeholder_node.target] = placeholder_node
        return Proxy(placeholder_node, self.tracer)

    def output(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``ouptput`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert isinstance(target, str)
        output_node = self.new_graph.output(target)
        output_node.args = (output_node.prev,)
        self.nodeDict[output_node.target] = output_node
        return Proxy(output_node, self.tracer)

    def get_attr(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``get_attr`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert isinstance(target, str)
        get_attr_node = self.new_graph.get_attr(target)
        self.nodeDict[get_attr_node.target] = get_attr_node
        return Proxy(get_attr_node, self.tracer)

    def call_function(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_function`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert callable(target)
        call_fnc_node = self.new_graph.call_function(target)
        call_fnc_node.args = process_args(args)
        call_fnc_node.kwargs = process_kwargs(kwargs)
        self.nodeDict["function"].append(call_fnc_node)
        return Proxy(call_fnc_node, self.tracer)

    def call_method(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_method`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert isinstance(target, str)
        call_method_node = self.new_graph.call_method(target)
        call_method_node.args = process_args(args)
        call_method_node.kwargs = process_kwargs(kwargs)
        self.nodeDict["method"].append(call_method_node)
        return Proxy(call_method_node, self.tracer)

    def call_module(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_module`` node. Adds mapping between target of the node and the node itself to self.nodeDict.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        assert isinstance(target, str)
        call_module_node = self.new_graph.call_module(target)
        call_module_node.args = process_args(args)
        call_module_node.kwargs = process_kwargs(kwargs)
        self.nodeDict["module"].append(call_module_node)
        return Proxy(call_module_node, self.tracer)

    def transform(self) -> dict:
        """
        Run `module` via interpretation and return the the nodeDict.

        Returns:
            A dictionary containing the mapping between the target of the node and the node itself.
        """
        result = super().run()
        if result is not None:

            def strip_proxy(a: Union[Argument, Proxy]) -> Any:
                return a.node if isinstance(a, Proxy) else a

            self.new_graph.output(map_aggregate(result, strip_proxy))
        return self.nodeDict



#!/usr/bin/env python3
import torch
import torch.fx as fx
from torch.fx.node import Argument, Node, Target
from typing import Any, Dict, Iterator, List, Optional, Tuple
from dmx.compressor import sparse, numerical
from dmx.compressor.functional.approximate import NoApproximation
from dmx.compressor.modeling import nn as dmxnn


class InfoNode:
    """Class for storing information of a node. args and kwargs attributes are used for Graphviz drawing"""

    def __init__(self, name: str, args: List, kwargs: Dict):
        self.name = name
        self.args = args
        self.kwargs = kwargs


class InfoEdge:
    """Class for storing information of an edge. args and kwargs attributes are used for Graphviz drawing"""

    def __init__(self, name: str, args: List, kwargs: Dict):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.device = None
        self.size = None
        self.dtype = None


class MetadataInterpreter(fx.Interpreter):
    """Interpreter that captures the metadata, eg: device, dtype, shape, of tensors that passed through the Graphmodule
    Example:
        >>> import dmx.compressor
        >>> import torch
        >>> net = torch.nn.Sequential(torch.nn.Linear(64,64),torch.nn.ReLU(),torch.nn.Linear(64,10))
        >>> gm = dmx.compressor.fx.transform.substitute_transform(net)
        >>> inp = torch.rand(1,64)
        >>> nodeDict = dmx.compressor.fx.transformer.NodeDictTransformer(gm).transform()
        >>> gi = dmx.compressor.utils.MetadataInterpreter(gm, nodeDict)
        >>> gi.nodes[0].__dict__ # Check node information
        {'name': '_0',
        'args': ['140117175215088', '_0_0'],
        'kwargs': {'fillcolor': '#D5D6EA', 'shape': 'circle'},
        'input_cast': SAME,
        'output_cast': SAME,
        'accum_cast': SAME,
        'weight_cast': SAME,
        'bias_cast': SAME,
        'weight_sparsifier': DENSE,
        'approximator': NONE}
        >>> e = [e for e in gi.edges]
        >>> e
        ['_0', 'input_1_0', '_1', '_2']
        >>> gi.edges['_0'].__dict__ # Check edge information
        {'name': '_0',
        'args': ['140117175215088',
        '140117175212144',
        '[1, 64]\ntorch.float32\ncpu\n'],
        'kwargs': {'arrowhead': 'open'},
        'size': [1, 64],
        'dtype': torch.float32,
        'device': device(type='cpu')}
    """

    def __init__(self, module: fx.GraphModule, nodeDict: Dict):
        super().__init__(module)

        self.nodeDict = nodeDict
        self.functionNum = 0
        self.methodNum = 0
        self.moduleNum = 0
        # A list that stores node information
        self.nodes = list()
        # A dictionary that stores edge info attached to the node, an edge is stored under the first node it attaches to
        # Both node and edge infos are stored as a tuple of args and kwargs
        self.edges = dict()
        self.input_names = []

    def run(self, *args, initial_env: Optional[Dict[Node, Any]] = None) -> Any:
        self.sizes = dict()
        self.dtypes = dict()
        self.devices = dict()
        self.env = initial_env if initial_env else {}

        # Positional function args are consumed left-to-right by
        # `placeholder` nodes. Use an iterator to keep track of
        # position and extract those values.
        self.args_iter: Iterator[Any] = iter(args)

        for node in self.module.graph.nodes:
            if node in self.env:
                # Short circuit if we have this value. This could
                # be used, for example, for partial evaluation
                # where the caller has pre-populated `env` with
                # values for a subset of the program.
                continue

            self.env[node] = self.run_node(node)
            if isinstance(self.env[node], torch.Tensor):
                self.sizes[node.name] = list(self.env[node].shape)
                self.dtypes[node.name] = self.env[node].dtype
                self.devices[node.name] = self.env[node].device
            elif isinstance(self.env[node], torch.Size):
                self.sizes[node.name] = list(self.env[node])
            if self.garbage_collect_values:
                for to_delete in self.user_to_last_uses.get(node, []):
                    del self.env[to_delete]

            if node.op == "output":
                output_val = self.env[node]
                return output_val

    def run_node(self, n: Node) -> Any:
        return super().run_node(n)

    def process_edge_metadata(self, edge_name: str, node_name: str) -> str:
        """Adds metadata of the edge to InfoEdge and generate the printout of the InfoEdge"""
        print_out = ""
        if node_name in self.sizes:
            self.edges[edge_name].size = self.sizes[node_name]
            print_out += str(self.edges[edge_name].size) + "\n"
        if node_name in self.dtypes:
            self.edges[edge_name].dtype = self.dtypes[node_name]
            print_out += str(self.edges[edge_name].dtype) + "\n"
        if node_name in self.devices:
            self.edges[edge_name].device = self.devices[node_name]
            print_out += str(self.edges[edge_name].device) + "\n"
        return print_out

    def add_edges(self, node: Node) -> None:
        target_name = node.name
        self.edges[target_name] = InfoNode(
            target_name,
            [str(id(target_name)), ""],
            {"arrowhead": "open"},
        )
        for i, argNode in enumerate(node.args):
            if isinstance(argNode, Node):
                # argNode is input
                if argNode.name in self.input_names:
                    name = argNode.name + target_name
                    self.edges[name] = InfoEdge(
                        name,
                        [str(id("start")), str(id(target_name)), argNode.name + ": "],
                        {"fillcolor": "green", "arrowsize": "2"},
                    )

                # argNode is input to multiple nodes
                elif len(self.edges[argNode.name].args) > 2:
                    name = argNode.name + target_name
                    self.edges[name] = InfoEdge(
                        name,
                        [str(id(argNode.name)), str(id(target_name)), ""],
                        {"arrowhead": "open"},
                    )

                else:
                    self.edges[argNode.name].args.insert(1, str(id(target_name)))
                    name = argNode.name
                print_out = self.process_edge_metadata(name, argNode.name)
                self.edges[name].args[-1] += print_out
            else:
                nodeName = repr(argNode) + target_name
                self.nodes.append(
                    InfoNode(
                        nodeName,
                        [str(id(nodeName)) + str(i), repr(argNode)],
                        {"fillcolor": "#D9E3DA", "shape": "oval"},
                    )
                )
                self.edges[nodeName + str(i)] = InfoEdge(
                    nodeName + str(i),
                    [str(id(nodeName)) + str(i), str(id(target_name)), ""],
                    {"arrowhead": "open"},
                )
        for kwargNode in node.kwargs.values():
            if isinstance(kwargNode, Node):
                # kwargNode is input to multiple nodes
                if len(self.edges[kwargNode.name].args) > 2:
                    name = kwargNode.name + target_name
                    self.edges[name] = InfoEdge(
                        name,
                        [str(id(kwargNode.name)), str(id(target_name)), ""],
                        {"arrowhead": "open"},
                    )
                else:
                    self.edges[kwargNode.name].args.insert(1, str(id(target_name)))
                    name = kwargNode.name
                print_out = self.process_edge_metadata(name, kwargNode.name)
                self.edges[name].args[-1] += print_out

    # Main Node running APIs
    def placeholder(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Green edge, target could be name
        self.input_names.append(self.nodeDict[target].name)
        return super().placeholder(target, args, kwargs)

    def get_attr(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Weight or bias (Square box)
        name = self.nodeDict[target].name
        self.nodes.append(
            InfoNode(name, [(str(id(name))), name], {"fillcolor": "#F9B5AC"})
        )
        self.edges[name] = InfoEdge(name, [str(id(name)), ""], {"arrowhead": "open"})
        return super().get_attr(target, args, kwargs)

    def call_function(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Diamond with target inside, different color for dmx ops
        # args are used to put text above the in edges
        # output is out edge
        call_func_node = self.nodeDict["function"][self.functionNum]
        self.functionNum += 1
        target_name = call_func_node.name
        self.nodes.append(
            InfoNode(
                target_name,
                [(str(id(target_name))), target_name],
                {"fillcolor": "#edc485", "shape": "diamond"},
            )
        )
        if "dmx" in repr(call_func_node.target):
            self.nodes[-1].kwargs["fillcolor"] = "#AAD8FF"
            self.nodes[-1].args[-1] = str(call_func_node.target)
        self.add_edges(call_func_node)
        return super().call_function(target, args, kwargs)

    def call_method(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Diamond with target inside (slightly different shade)
        # args are used to put text above the in edges
        # output is out edge
        call_method_node = self.nodeDict["method"][self.methodNum]
        self.methodNum += 1
        target_name = call_method_node.name
        self.nodes.append(
            InfoNode(
                target_name,
                [(str(id(target_name))), target_name],
                {"fillcolor": "#F3DCD4", "shape": "diamond"},
            )
        )
        self.add_edges(call_method_node)
        return super().call_method(target, args, kwargs)

    def call_module(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Circle with target inside
        # args are used to put text above the in edges
        # output is out edge

        color = "#f0f8ff"
        call_module_node = self.nodeDict["module"][self.moduleNum]
        self.moduleNum += 1
        curr_mod = self.module.get_submodule(target)
        target_name = call_module_node.name
        print_out = target + "\n" + f"{curr_mod.__class__}"
        self.nodes.append(
            InfoNode(
                target_name,
                [(str(id(target_name))), ""],
                {"fillcolor": color, "shape": "circle"},
            )
        )
        if isinstance(curr_mod, dmxnn.DmxModule):
            self.nodes[-1].kwargs["fillcolor"] = "#D5D6EA"
            for op_name, cast in curr_mod.named_modules():
                if isinstance(cast, numerical.CastTo):
                    format = cast.format
                    setattr(self.nodes[-1], op_name, format)
                    if not isinstance(format, numerical.Same):
                        print_out += "\n" + f"{op_name}: {repr(format)}"

            sparsifier = curr_mod.weight_sparsifier
            if sparsifier:
                sparseness = sparsifier.sparseness
                setattr(self.nodes[-1], "weight_sparsifier", sparseness)
                if not isinstance(sparseness, sparse.Dense):
                    print_out += "\n" + f"weight_sparsifier: {repr(sparseness)}"

            approximator = curr_mod.approximator
            if approximator:
                approx_func = approximator.function
                setattr(self.nodes[-1], "approximator", approx_func)
                if not isinstance(approx_func, NoApproximation):
                    print_out += "\n" + f"approximator: {repr(approx_func)}"
        self.nodes[-1].args[-1] += print_out
        self.add_edges(call_module_node)
        return super().call_module(target, args, kwargs)

    def output(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        # Blue arrow with output text
        output_node = self.nodeDict[target]
        for argNode in output_node.args:
            self.edges[argNode.name].args.insert(1, str(id("end")))
            self.edges[argNode.name].kwargs["arrowhead"] = "normal"
            self.edges[argNode.name].kwargs["arrowsize"] = "2"
            self.edges[argNode.name].kwargs["fillcolor"] = "blue"
            self.edges[argNode.name].args[2] = "output: " + self.process_edge_metadata(
                argNode.name, argNode.name
            )
        return super().output(target, args, kwargs)



import pytest
import torch
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor import format
from dmx.compressor.modeling.nn.experimental import (
    Conv1dScatter,
    Conv1dUnfold,
    Conv2dGather,
    Conv2dUnfold,
)


RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize("bias", (True, False))
@pytest.mark.parametrize(
    "in_channels,out_channels,kernel_size,stride",
    (
        (1, 8, 5, 1),
        (3, 16, 5, 1),
        (3, 16, 3, 1),
        (64, 64, 3, 1),
        (64, 128, 3, 2),
    ),
)
@pytest.mark.parametrize(
    "image_size",
    (
        (28, 28),
        (32, 32),
        (224, 224),
    ),
)
@pytest.mark.parametrize(
    "batch_size",
    (
        1,
        16,
    ),
)
@pytest.mark.parametrize(
    "module_type",
    (
        "Conv2d",
        "ConvTranspose2d",
    ),
)
def test_conv2d(
    module_type,
    batch_size,
    image_size,
    in_channels,
    out_channels,
    kernel_size,
    stride,
    bias,
):
    torch_module = eval(f"torch.nn.{module_type}")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module = eval(f"dmxnn.{module_type}")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module.weight.data = torch_module.weight.data
    if bias:
        dmx_module.bias.data = torch_module.bias.data
    t_inp = torch.randn(
        batch_size, in_channels, *image_size, device=device
    ).requires_grad_()
    c_inp = t_inp.clone().detach().requires_grad_()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    g_out = torch.randn_like(t_out)
    t_out.backward(g_out)
    c_out.backward(g_out)
    assert torch.allclose(t_out.data, c_out.data, atol=1e-6)
    assert torch.allclose(t_inp.grad, c_inp.grad, atol=1e-6)


@pytest.mark.parametrize("bias", (True, False))
@pytest.mark.parametrize(
    "in_channels,out_channels,kernel_size,stride",
    (
        (1, 8, 5, 1),
        (3, 16, 5, 1),
        (3, 16, 3, 1),
        (64, 64, 3, 1),
        (64, 128, 3, 2),
    ),
)
@pytest.mark.parametrize(
    "batch_size",
    (
        1,
        16,
    ),
)
@pytest.mark.parametrize("image_size_1d", (28, 32, 224))
def test_conv1d(
    batch_size, image_size_1d, in_channels, out_channels, kernel_size, stride, bias
):
    torch_module = eval(f"torch.nn.Conv1d")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module = eval(f"dmxnn.Conv1d")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module.weight.data = torch_module.weight.data
    if bias:
        dmx_module.bias.data = torch_module.bias.data
    t_inp = torch.randn(
        batch_size, in_channels, image_size_1d, device=device
    ).requires_grad_()
    c_inp = t_inp.clone().detach().requires_grad_()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    g_out = torch.randn_like(t_out)
    t_out.backward(g_out)
    c_out.backward(g_out)
    assert torch.allclose(t_out.data, c_out.data, atol=1e-6)
    assert torch.allclose(t_inp.grad, c_inp.grad, atol=1e-6)


@pytest.mark.parametrize("bias", (True, False))
@pytest.mark.parametrize(
    "in_channels,out_channels,kernel_size,stride",
    (
        (80, 384, 3, 1),
        (384, 384, 3, 2),
    ),
)
@pytest.mark.parametrize(
    "batch_size",
    (1,),
)
@pytest.mark.parametrize("image_size_1d", (30,))
def test_conv1d_scatter(
    batch_size, image_size_1d, in_channels, out_channels, kernel_size, stride, bias
):
    torch_module = eval(f"torch.nn.Conv1d")(
        in_channels,
        out_channels,
        kernel_size,
        stride=stride,
        bias=bias,
        device=device,
    )
    dmx_module = eval(f"dmxnn.Conv1d")(
        in_channels,
        out_channels,
        kernel_size,
        stride=stride,
        bias=bias,
        device=device,
    )
    dmx_module_scatter = eval(f"Conv1dScatter")(
        in_channels,
        out_channels,
        kernel_size,
        stride=stride,
        bias=bias,
        device=device,
    )
    dmx_module_unfold = eval(f"Conv1dUnfold")(
        in_channels,
        out_channels,
        kernel_size,
        stride=stride,
        bias=bias,
        device=device,
    )
    dmx_module.weight.data = torch_module.weight.data
    dmx_module_scatter.weight.data = torch_module.weight.data
    dmx_module_unfold.weight.data = torch_module.weight.data

    if bias:
        dmx_module.bias.data = torch_module.bias.data
        dmx_module_scatter.bias.data = torch_module.bias.data
        dmx_module_unfold.bias.data = torch_module.bias.data

    ### testing to_compiler_graph
    scatter_qdq = torch.fx.GraphModule(
        dmx_module_scatter, dmx_module_scatter.to_compiler_graph()
    )
    unfold_qdq = torch.fx.GraphModule(
        dmx_module_unfold, dmx_module_unfold.to_compiler_graph()
    )

    t_inp = torch.randn(
        batch_size, in_channels, image_size_1d, device=device
    ).requires_grad_()
    c_inp = t_inp.clone().detach().requires_grad_()
    c_inp_scatter = t_inp.clone().detach().requires_grad_()
    c_inp_unfold = t_inp.clone().detach().requires_grad_()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    c_out_scatter = dmx_module_scatter(c_inp_scatter)
    c_out_unfold = dmx_module_unfold(c_inp_unfold)
    c_out_qdq_scatter = scatter_qdq(c_inp_scatter.clone().detach())
    c_out_qdq_unfold = unfold_qdq(c_inp_unfold.clone().detach())

    g_out = torch.randn_like(t_out)
    t_out.backward(g_out)
    c_out.backward(g_out)
    c_out_scatter.backward(g_out)
    c_out_unfold.backward(g_out)
    assert torch.allclose(c_out.data, c_out_scatter.data, atol=1e-5)
    assert torch.allclose(t_out.data, c_out_scatter.data, atol=1e-5)
    assert torch.allclose(t_out.data, c_out_unfold.data, atol=1e-5)
    assert torch.allclose(c_out_qdq_unfold.data, c_out_unfold.data, atol=1e-5)
    assert torch.allclose(c_out_qdq_scatter.data, c_out_scatter.data, atol=1e-5)

    assert torch.allclose(c_inp.grad, c_inp_scatter.grad, atol=1e-5)
    assert torch.allclose(t_inp.grad, c_inp_scatter.grad, atol=1e-5)
    assert torch.allclose(t_inp.grad, c_inp_unfold.grad, atol=1e-5)

    # config rule corresponds to the Conv1d config in the BASIC mode
    config_rule = dict(
        input_formats=[format.BFP16_64],
        weight_format=format.BFP16_64,
        bias_format=format.BFP32_1,
        output_formats=[format.FLOAT16],
    )
    dmx_module.configure(config_rule)
    dmx_module_scatter.configure(config_rule)
    dmx_module_unfold.configure(config_rule)
    c_out = dmx_module(c_inp)
    c_out_scatter = dmx_module_scatter(c_inp_scatter)
    c_out_unfold = dmx_module_unfold(c_inp_unfold)
    g_out = torch.randn_like(t_out)
    c_out.backward(g_out)
    c_out_scatter.backward(g_out)
    c_out_unfold.backward(g_out)
    assert torch.allclose(c_out.data, c_out_scatter.data, atol=1e-5)
    assert torch.allclose(c_inp.grad, c_inp_scatter.grad, atol=1e-5)
    assert torch.allclose(c_out.data, c_out_unfold.data, atol=1e-5)
    assert torch.allclose(c_inp.grad, c_inp_unfold.grad, atol=1e-5)


@pytest.mark.parametrize("bias", (False,))
@pytest.mark.parametrize(
    "in_channels,out_channels,kernel_size,stride",
    (
        (3, 768, 16, 16),
        (3, 768, 32, 32),
        (3, 1024, 14, 14),
    ),
)
@pytest.mark.parametrize(
    "image_size",
    ((224, 224),),
)
@pytest.mark.parametrize(
    "batch_size",
    (1,),
)
def test_conv2d_gather(
    batch_size,
    image_size,
    in_channels,
    out_channels,
    kernel_size,
    stride,
    bias,
):
    torch_module = eval(f"torch.nn.Conv2d")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module = eval(f"dmxnn.Conv2d")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module_gather = eval(f"Conv2dGather")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module_unfold = eval(f"Conv2dUnfold")(
        in_channels, out_channels, kernel_size, stride=stride, bias=bias, device=device
    )
    dmx_module.weight.data = torch_module.weight.data
    dmx_module_gather.weight.data = torch_module.weight.data
    dmx_module_unfold.weight.data = torch_module.weight.data

    if bias:
        dmx_module.bias.data = torch_module.bias.data
        dmx_module_gather.bias.data = torch_module.bias.data
        dmx_module_unfold.bias.data = torch_module.bias.data

    t_inp = torch.randn(
        batch_size, in_channels, *image_size, device=device
    ).requires_grad_()
    c_inp = t_inp.clone().detach().requires_grad_()
    c_inp_gather = t_inp.clone().detach().requires_grad_()
    c_inp_unfold = t_inp.clone().detach().requires_grad_()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    c_out_gather = dmx_module_gather(c_inp_gather)
    c_out_unfold = dmx_module_unfold(c_inp_unfold)

    g_out = torch.randn_like(t_out)
    t_out.backward(g_out)
    c_out.backward(g_out)
    c_out_gather.backward(g_out)
    c_out_unfold.backward(g_out)
    assert torch.allclose(c_out.data, c_out_gather.data, atol=1e-5)
    assert torch.allclose(t_out.data, c_out_gather.data, atol=1e-5)
    assert torch.allclose(t_out.data, c_out_unfold.data, atol=1e-5)

    assert torch.allclose(c_inp.grad, c_inp_gather.grad, atol=1e-5)
    assert torch.allclose(t_inp.grad, c_inp_gather.grad, atol=1e-5)
    assert torch.allclose(t_inp.grad, c_inp_unfold.grad, atol=1e-5)

    # config rule corresponds to the Conv2d config in the BASIC mode
    config_rule = dict(
        input_formats=[format.BFP16_64],
        weight_format=format.BFP16_64,
        bias_format=format.BFP32_1,
        output_formats=[format.FLOAT16],
    )
    dmx_module.configure(config_rule)
    dmx_module_gather.configure(config_rule)
    dmx_module_unfold.configure(config_rule)
    c_out = dmx_module(c_inp)
    c_out_gather = dmx_module_gather(c_inp_gather)
    c_out_unfold = dmx_module_unfold(c_inp_unfold)
    g_out = torch.randn_like(t_out)
    c_out.backward(g_out)
    c_out_gather.backward(g_out)
    c_out_unfold.backward(g_out)
    assert torch.allclose(c_out.data, c_out_gather.data, atol=1e-5)
    assert torch.allclose(c_inp.grad, c_inp_gather.grad, atol=1e-5)
    assert torch.allclose(c_out.data, c_out_unfold.data, atol=1e-5)
    assert torch.allclose(c_inp.grad, c_inp_unfold.grad, atol=1e-5)



import math
import torch
import pptree
import dmx.compressor


def mask2braille(m, dims=(0, 1), max_elems=None):
    # permute dimensions
    ds = list(range(m.dim()))
    ds[0], ds[dims[0]] = dims[0], ds[0]
    ds[1], ds[dims[1]] = dims[1], ds[1]
    m = m.permute(*ds)
    truncate_h = truncate_w = False
    if max_elems is not None:
        truncate_h = max_elems < m.size(0)
        truncate_w = max_elems < m.size(1)
        m = m[:max_elems, :max_elems, ...]
    # reduce mask
    _h, _w = m.size(0), m.size(1)
    # m = m.reshape(_h, _w, -1).sum(-1) > 0
    m = m.reshape(_h, _w, -1)[:, :, 0]
    # pad mask
    h, w = math.ceil(_h / 4) * 4, math.ceil(_w / 2) * 2
    m = torch.nn.functional.pad(m, (0, w - _w, 0, h - _h))
    # compute values
    k = torch.Tensor([[0, 3], [1, 4], [2, 5], [6, 7]])
    x = (
        torch.nn.functional.conv2d(
            m.unsqueeze(0).unsqueeze(0).float(),
            2.0 ** k.to(m.device).unsqueeze(0).unsqueeze(0),
            stride=(4, 2),
        )
        .squeeze(0)
        .squeeze(0)
        .int()
        + 10240
    )
    x = "\n".join(["".join([chr(_x) for _x in row]) for row in x])
    return _box_wrap(
        x, pad_left=False, pad_right=truncate_w, pad_top=False, pad_bottom=truncate_h
    )


def _box_wrap(
    mls: str,
    pad_left: bool = False,
    pad_right: bool = False,
    pad_top: bool = False,
    pad_bottom: bool = False,
) -> str:
    lines = mls.split("\n")
    pad_char = "\u2592"
    if pad_left:
        lines = [pad_char + ln for ln in lines]
    if pad_right:
        lines = [ln + pad_char for ln in lines]
    if pad_top:
        lines = [pad_char * len(lines[0])] + lines
    if pad_bottom:
        lines = lines + [pad_char * len(lines[-1])]
    nchar = len(lines[0])
    lines = ["\u2502" + ln + "\u2502" for ln in lines]
    lines = (
        ["\u256d" + "\u2500" * nchar + "\u256e"]
        + lines
        + ["\u2570" + "\u2500" * nchar + "\u256f"]
    )
    return "\n".join(lines)


def print_model_tree(model: torch.nn.Module, include_type=False) -> str:
    def node_str(_n, _m):
        return f"{_n}:{type(_m).__name__}" if include_type else _n

    is_leaf_node = dmx.compressor.nn.is_configurable
    m_root = pptree.Node(node_str("_gm", model._gm))

    def get_tree(mod, root):
        for n, m in mod.named_children():
            n_ = pptree.Node(node_str(n, m), root)
            if not is_leaf_node(m):
                get_tree(m, n_)

    get_tree(model._gm, m_root)
    pptree.print_tree(m_root)



import pytest
import torch
import torch.nn as nn
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor.advanced_recipe import DmxModuleSmoothQuantHyperparams


RANDOM_SEED = 0

torch.manual_seed(RANDOM_SEED)


BATCH_SIZE = 2
IN_DIM = 16
OUT_DIM = 8
PERTURB_CHAN = 4
KER_SIZE = 3
IMG_SIZE = 8


def _create_module(cls):
    if cls == dmxnn.Linear:
        return cls(IN_DIM, OUT_DIM)
    elif cls == dmxnn.Conv1d:
        return cls(IN_DIM, 2, KER_SIZE)
    elif cls == dmxnn.Conv2d:
        return cls(IN_DIM, OUT_DIM, KER_SIZE, KER_SIZE)
    else:
        raise ValueError("unsupported module class {cls}")


def _create_test_input(module: torch.nn.Module, scaler: float):
    with torch.no_grad():
        if isinstance(module, nn.Conv1d):
            x_ref = torch.randn(BATCH_SIZE, IN_DIM, IMG_SIZE)
        elif isinstance(module, nn.Conv2d):
            x_ref = torch.randn(BATCH_SIZE, IN_DIM, IMG_SIZE, IMG_SIZE)
        else:
            x_ref = torch.randn(BATCH_SIZE, IN_DIM)

        index_shape = [1] * x_ref.ndim
        index_shape[module.ch_axis] = PERTURB_CHAN
        index = torch.arange(PERTURB_CHAN).reshape(index_shape)

        # Perturb the values at the specified indices
        perturb_slice = [slice(None)] * x_ref.ndim
        perturb_slice[module.ch_axis] = index
        x_ref[perturb_slice] *= scaler

    return x_ref


@pytest.mark.parametrize(
    "module_cls",
    (
        dmxnn.Linear,
        dmxnn.Conv1d,
        dmxnn.Conv2d,
    ),
)
@pytest.mark.parametrize("dynamic", (True, False))
@pytest.mark.parametrize("migration_strength", (0.5, 0.0, 1.0))
@pytest.mark.parametrize("scale_format", ("SAME", "FP[0|8|0,127](FN)"))
@pytest.mark.parametrize("perturbation_scaler", (1.0, 1e-6, 1e6))
def test_smoothquant(
    module_cls, dynamic, migration_strength, scale_format, perturbation_scaler
):
    module = _create_module(module_cls)
    assert module.smoothquant.enabled[0] == 0  # disabled

    module.smoothquant.set_dynamic(dynamic)
    module.smoothquant.set_migration_strength(migration_strength)
    module.smoothquant.set_scale_format(scale_format)

    x_ref = _create_test_input(module, perturbation_scaler).requires_grad_()
    x = x_ref.clone().detach().requires_grad_()
    y_ref = module(x_ref)
    g_out = torch.randn_like(y_ref)
    y_ref.backward(g_out)
    x_ref_grad = x_ref.grad
    w_ref_grad = module.weight.grad
    w_ref = module.weight.data

    module.zero_grad()
    module.smoothquant.enable()
    if not dynamic:  # static mode needs calibration
        with module.calibrating_smoothquant(
            DmxModuleSmoothQuantHyperparams(migration_strength=migration_strength)
        ):
            module(x)

    y = module(x)
    y.backward(g_out)
    x_grad = x.grad
    w_grad = module.weight.grad
    w_scaled = module.smoothquant.scale_weight(module.weight)
    w = torch.div(w_scaled.movedim(module.win_ch_axis, -1), module.smoothquant.scale)
    w = w.movedim(-1, module.win_ch_axis)

    assert torch.allclose(y, y_ref, atol=1e-6, rtol=1e-3)
    assert torch.allclose(x_grad, x_ref_grad, atol=1e-6, rtol=1e-3)
    assert torch.allclose(w_grad, w_ref_grad, atol=1e-6, rtol=1e-3)
    assert torch.allclose(w, w_ref)



import torch
from torch import nn


def test_sigmoid_gradients():
    """Sigmoid function itself has no gradient problems."""
    score = nn.Parameter(torch.Tensor(5, 5))
    mask = torch.sigmoid(score)
    loss = torch.sum(mask)
    loss.backward()
    assert score.grad is not None


def test_bernoulli_gradients():
    """Bernoulli sampler introduces gradient problem."""
    score = nn.Parameter(torch.Tensor(5, 5))
    mask = torch.sigmoid(score)
    mask = torch.bernoulli(mask)
    loss = torch.sum(mask)
    loss.backward()
    assert (score.grad is None) or (torch.sum(score.grad) == 0)
